/**
 * REST demo programs
 */
package demo.rest;

//HTTP Classes
import org.apache.wink.client.ClientConfig;
import org.apache.wink.client.ClientWebException;
import org.apache.wink.client.RestClient;
import org.apache.wink.client.Resource;
import org.apache.wink.client.ClientResponse;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.lang.StringUtils;      

// Utility Classes
import java.io.InputStream;
import java.io.ByteArrayInputStream;
import java.io.UnsupportedEncodingException;

import java.util.List;
import javax.ws.rs.core.MultivaluedMap; 

/**
 * A base client that handles login and logout, formats messages, and 
 * provides utility methods for programs that use REST APIs. 
 * This program makes use of demo.rest.BaseUtil.java.
 * <p>
 * Note:<br/>
 * This class uses the Apache wink RestClient, which makes it
 * pretty easy to make requests and handle responses. 
 * 
 * TODO: Make the program and programs extended from it testable:
 *   --All functions return a string or param-list with a value or an error
 *   --Value-producing functions never do output
 *   --Utility functions that generate output never return values
 *      
 * @author Eric Armstrong      
 */
public class BaseClient {
  
  static String cookieString;
  private static RestClient client;
  
  static String baseUrl;
  static String login_url;
  static String logout_url;
  static String username; 
  static String password; 

  /**
   * Log in to the platform, giving username and password. 
   * 
   * @return a String containing cookie parameters required for subsequent
   *         REST requests, including the session ID and the load balancer
   *         key (if any) to ensure that a session-based load balancer 
   *         routes  subsequent requests to the same server.
   */
  public String login()
  { 
    // TODO: Read baseURL, username, and password from a file, so the
    // program can be run from different locations without modifications.
    baseUrl = "http://{domain}/networking/rest";
    username = "yourName@yourEmailAddress"; 
    password = "yourPassword"; 
    
    login_url = baseUrl + "/login";
    logout_url = baseUrl + "/logout";
    String xml = "<platform>"
        + "<login>"
            + "<userName>"+username+"</userName>"
            + "<password>"+password+"</password>"
        + "</login>"
        + "</platform>";
    try
    {
      System.out.println("Logging in");
      ClientConfig config = new ClientConfig();
      config.connectTimeout(60000); // 1 minute
      config.readTimeout(600000);  // 10 minutes
      client = new RestClient(config);

      Resource resource = client.resource(login_url);
      resource.contentType("application/xml");
      resource.accept("application/xml");
      ClientResponse response = resource.post(xml);
      
      // Combine cookie parameters from the response header. 
      // Make a cookie string for use in subsequent requests.
      MultivaluedMap<String,String> headers = response.getHeaders();  
      //BaseUtil.prettyPrint(headers);
      String cookieString = "";
      List<String> cookieParams = headers.get("Set-Cookie");
      cookieString = StringUtils.join(cookieParams, "; ");

      return cookieString;
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
    return null;
  }
  
  /**
   * Use the connection created by the login() method to get a Resource
   * object specified by a universal resource identifier (URI).
   *  
   * @param uri
   * @return the resource named by the uri
   * @throws Exception 
   */
  public static Resource getResource(String uri) throws Exception
  {
      if (client == null) throw new Exception(
          "Successful login() required to get a Resource");
      return client.resource(uri);
  }

  /**
   * Get a value from XML returned in a response.
   * 
   * @param xpath_expr  An XPATH expression. 
   *                    For example: "/platform/login/sessionId"
   *                    
   * @param xml The XML string to be parsed. Generally obtained using
   *        ClientResponse.getEntity(String.class) --Apache wink RestClient
   *        httpMethod.getResponseBodyAsString()   --Apache HttpClient
   */
  public static String getValue(String xpath_expr, String xml) 
  { 
    return BaseUtil.xpath_result(xpath_expr, xml);
  }
  
  /**
   * Echo the response in pretty-printed form.
   * 
   * @param xml_in An XML InputStream, generally obtained using
   *            httpMethod.getResponseBodyAsStream
   */
  public static void echoResponse(InputStream xml_in)
  {
    BaseUtil.prettyPrint(xml_in);
  }
  
  /**
   * Echo a response that is an XML string
   * @param xml The XML string to echo
   */
  public static void echoResponse(String xml)
  {
      try {
        InputStream xml_in = new ByteArrayInputStream(xml.getBytes("UTF-8"));
        echoResponse(xml_in);
      }
      catch (UnsupportedEncodingException e) {
          e.printStackTrace();
      }
  }
  
  /**
   * Echo the response returned in a web exception.
   * <p>
   * This method handles the unclosed horizontal-rule elements (<HR>) in 
   * the HTML error-pages generated by Apache Tomcat, so XML parsing can be
   * used to pretty print the error page, after a fashion. 
   *    
   * @param webException The exception returned by a REST call
   */
  public static void echoResponse(ClientWebException webException)
  {
      String response = webException.getResponse().getEntity(String.class);
      StringBuffer sb = new StringBuffer(response);
      //System.out.println(response);
      int i = sb.indexOf("<HR");
      while (i > 0) 
      {
        int j = sb.indexOf(">",i+1);
        sb.insert(j,'/');
        i = sb.indexOf("<HR",j+1);
     }
     //System.out.println(sb.toString());       
     echoResponse(sb.toString());
  }
  
  /**
   * Echo the status of the request.
   * 
   * @param status  Status code returned by httpClient.executeMethod()
   * @param responseXml String containing XML returned in the response,
   *        generally obtained using httpMethod.getResponseBodyAsString()
   */
  public void echoStatus(int status, String xml) {
    System.out.println("HTTP error status is "+ status);
    System.out.println(statusMessage(status));
    String code = getValue("platform/message/code", xml);
    String msg  = getValue("platform/message/description", xml);
    System.out.println("  Platform message code is " + code);
    System.out.println("  Platform message is:\n    " + msg);
  }
   
  /**
   * Return message text for HTTP status codes.
   * 
   * @param status  The status code returned by the request.
   * @param responseXml  The XML string returned by the request, 
   *        generally obtained using httpMethod.getResponseBodyAsString() 
   */
  public String statusMessage(int status) 
  {
    String result = "";
    switch(status) {
      case 200: 
        result = "SUCCESS\n  "
          + "The request succeeded.";
        break;
      case 201: 
        result = "CREATE SUCCEEDED\n  "
          + "A new entity was successfully created.";
        break;
      case 400: 
        result = "BAD_REQUEST\n  "
          + "Invalid argument, badly-formed request, or request rejected.\n  "
          + "Check content types, data format, and platform message code.";
        break;
      case 404: 
        result = "BAD_URL\n  "
          + "The resource does not exist or the server could not be reached.";
        break;
      case 405:
        result = "METHOD_NOT_ALLOWED\n  "
          + "This method (get/post/put/delete) is not implemented at this URI";
        break;
      case 415: 
        result = "UNSUPPORTED_MEDIA_TYPE\n  "
          + "Unknown or unspecified Content-type specified in the request header.";
        break;
      case 999: 
        result = "tbd\n  "
               + "tbd.";
        break;
      default:
        result = HttpStatus.getStatusText(status);
        break;
    }
    return result;
  }

  /**
   * Terminate the session.
   */
  public void logout()
  {
    Resource resource = client.resource(logout_url);
    resource.header("Cookie", cookieString);                   // <---- CHANGE
    resource.accept("application/xml");
    ClientResponse response = resource.get();
    //String responseXml = response.getEntity(String.class);    
    //echoResponse(responseXml);
    System.out.println("Logged out");
  }
  
  /**
   * Test login capability and illustrate usage.
   */                                                             
  public static void main(String args[])
  {
    BaseClient client = new BaseClient();
    String cookieString = client.login();                      
    System.out.println("Session cookie:\n  " + cookieString);  
    try {
       // Utilize a resource
       String resourceUrl = baseUrl + "/user/info";  // Current user info
       Resource resource = client.getResource(resourceUrl); 
       resource.header("Cookie", cookieString);
       resource.accept("application/xml");              // or application/json
       ClientResponse response = resource.get();        // Fetch data
       // ClientResponse response = resource.post(xml); // Add data
       // ClientResponse response = resource.put(xml);  // Update data

       // Inspect the response
       String responseXml = response.getEntity(String.class);
       echoResponse(responseXml);        
    }    
    catch (ClientWebException webException)
    {
        echoResponse(webException);
    }
    catch (Exception e)
    {
        System.out.println(e.getMessage());
        e.printStackTrace();
    }
    finally
    {
        client.logout();
    }
  }
}
